from itertools import combinations, combinations_with_replacement, product
from fractions import Fraction


def compute_realization_geometric(perm_triplet):
    """Check if the triplet is realizable, possibly by reversing some permutations"""
    px, py, pz = perm_triplet
    rx, ry, rz = px[::-1], py[::-1], pz[::-1]
    for triplet in product([px, rx], [py, ry], [pz, rz]):
        r = compute_realization_canonical(triplet)
        if r is not None:
            return r
    return r


def compute_realization_canonical(perm_triplet):
    """Search for a triplet of tags that makes the permutations canonically
    realizable. Output the realization if it works, None otherwise"""
    n = len(perm_triplet[0])
    L = [list(x) for x in combinations_with_replacement(range(n + 1), 2)]
    nb = 0
    for ZOx, ZOy, ZOz in product(L, L, L):
        tag_triplet = [ZOx, ZOy, ZOz]
        if contains_forbidden_pattern(perm_triplet, tag_triplet):
            continue
        r = compute_realization_tagged(perm_triplet, tag_triplet)
        if r is not None:
            return r
    return None


def compute_realization_tagged(perm_triplet, tag_triplet):
    """Computes a realization of a tagged-perm triplet if it is possible and returns None otherwise
    perm_triplet is a list of 3 size-n permutations, each encoded as a list [2,1,0] for p(0)=2, p(1)=1 and p(2)=0
    tag_triplet is a list of 3 tags, each encoded as a pair [z,o] where z is the position of the 0 and o of the 1
    if | represents the positions of the tags (in order), 2|01| is encoded by its permutation [2,0,1] and its tag [0,3]
    """
    Px, Py, Pz = perm_triplet
    n = len(Px)
    Ix, Iy, Iz = inverse_perm(Px), inverse_perm(Py), inverse_perm(Pz)  # compute the inverses once for all
    inv_triplet = [Ix, Iy, Iz]

    # From now on, the vertices x_i, y_j, z_k are encoded by the numbers i, n+j and 2n+k respectively

    # pre-compute the interval of each vertex:
    regions = [compute_region(v, inv_triplet, tag_triplet) for v in range(3 * n)]

    # initialization of the adjacence matrix of the graph M[v][w]=True iff v->w
    M = [[False] * (3 * n) for _ in range(3 * n)]

    # the base matrix containing information from the perms and tags only, which will not change during the process
    B = compute_base_graph(perm_triplet, tag_triplet)

    # SV is a map, s.t. SV[(i,j)] is the sign vector of (i,j)
    # a sign vector is a 6-tuple containing pairs of the form (sign, unknown)
    # where sign = +- 1 and unknown = None if the sign is determined
    # and unknown = v if the value is the sign of sign*(v-f(w)), which is not known atm
    # U is the set of unknown and required v-f(w), identified by their v (w is associated_vertex(v))
    SV, U = {}, set([])
    for i, j in combinations(range(n), 2):  # for every pair of triangles
        SV[(i, j)] = list()
        for l in range(3):
            SV[(i, j)].append(compute_initial_orientation(i, l, j, inv_triplet, regions, U))
        for l in range(3):
            SV[(i, j)].append(compute_initial_orientation(j, l, i, inv_triplet, regions, U))
    m = len(U)

    for it in range(2 ** m):  # main loop: iteration over the orientations of the elements of U
        for ii, jj in product(range(3 * n), range(3 * n)):  # copy the base B into M
            M[ii][jj] = B[ii][jj]
        # if the i-th bit of it is 1 then v < f(w) for the i-th v of U and v > f(w) otherwise
        j = it
        for v in U:
            w = associated_vertex(v, n)
            if j % 2 == 1:
                M[v][w] = True
            else:
                M[w][v] = True
            j //= 2
        if not transitive_acyclic(M):
            # the graph is not acyclic, we try the next orientation of edges in U, as this one does not work
            continue
        for i, j in combinations(range(n), 2):  # check whether they are pairwise disjoint
            intersection = False
            # compute the sign vector of (i,j), now that we have all the required information
            sv = compute_sign_vector(SV[(i, j)], M)

            # we now apply the Devillers-Guigue algorithm
            if sv[0] == sv[1] == sv[2] or sv[3] == sv[4] == sv[5]:
                # Ti and Tj are disjoint, continue to the next pair
                continue
            Xi, Yi, Zi, Xj, Yj, Zj = i, i + n, i + 2 * n, j, j + n, j + 2 * n  # our encoding
            sign_i, sign_j = sv[0], sv[3]
            if sv[0] == sv[1]:
                Xj, Yj, Zj = Zj, Xj, Yj
                sign_i = sv[2]
            elif sv[0] == sv[2]:
                Xj, Yj, Zj = Yj, Zj, Xj
                sign_i = sv[1]
            if sv[3] == sv[4]:
                Xi, Yi, Zi = Zi, Xi, Yi
                sign_j = sv[5]
            elif sv[3] == sv[5]:
                Xi, Yi, Zi = Yi, Zi, Xi
                sign_j = sv[4]
            if sign_i == -1:
                Yi, Zi = Zi, Yi
            if sign_j == -1:
                Yj, Zj = Zj, Yj
            # the points have been permuted so that their sign vector is (1,-1,-1,1,-1,-1)
            # we (try to) compute the two orientations of the algorithm
            # each O is either
            O1 = compute_last([Xi, Yi, Xj, Yj], M, inv_triplet, regions)
            O2 = compute_last([Xi, Zi, Zj, Xj], M, inv_triplet, regions)
            if O1[1] is not None and O2[1] is not None:  # for some reason, this case never happen
                raise Exception("Both orientations are unknown", i, j, it)
            if O1[1] is None and O1[0] == 1:
                continue  # the two triangles are disjoint
            if O2[1] is None and O2[0] == 1:
                continue  # the two triangles are disjoint
            if O1[1] is None and O2[1] is None:  # they are both -1 at this point => Ti and Tj intersect
                intersection = True
                break  # the (i,j) loop
            if O1[1] is None:  # O1 is -1 and O2 is unknown: it has to be 1 otherwise the triangles intersect
                v, w = O2[1]
                if O2[0] == 1:
                    add_edge_closure(M, w, v)  # w -> v so that v - f(w) > 0
                else:
                    add_edge_closure(M, v, w)  # v -> w so that v - f(w) < 0
            if O2[1] is None:  # O2 is -1 and O1 is unknown: it has to be 1 otherwise the triangles intersect
                v, w = O1[1]
                if O1[0] == 1:
                    add_edge_closure(M, w, v)  # w -> v so that v - f(w) > 0
                else:
                    add_edge_closure(M, v, w)  # v -> w so that v - f(w) < 0
        if intersection:
            continue
        # at this stage a realization is possible
        L = topological_sort(M)
        return output_realization(L, regions, n)
    return None


def contains_forbidden_pattern(perm_triplet, tag_triplet):
    return contains_forbidden2_coded(perm_triplet, tag_triplet) or contains_forbidden3_coded(perm_triplet,
                                                                                             tag_triplet) or contains_forbidden4_coded(
        perm_triplet, tag_triplet)


def contains_forbidden2_coded(perm_triplet, tag_triplet):
    """Check whether the tagged triplet contains a minimal forbidden pattern of length 2"""
    if len(perm_triplet[0]) < 2:
        return False
    P, T = perm_triplet, tag_triplet
    I = [inverse_perm(p) for p in P]
    for _ in range(3):
        for i, j in combinations(range(len(P[0])), 2):
            if I[0][i] < I[0][j]:
                a, b = i, j
            else:
                a, b = j, i
            # ab0 1ab ab
            if I[0][b] < T[0][0] and T[1][1] <= I[1][a] < I[1][b] and I[2][a] < I[2][b]:
                return True
            # 0ab ab1 ba
            if T[0][0] <= I[0][a] and I[1][a] < I[1][b] < T[1][1] and I[2][b] < I[2][a]:
                return True
            # ab0 ba1 ab
            if I[0][b] < T[0][0] and I[1][b] < I[1][a] < T[1][1] and I[2][a] < I[2][b]:
                return True
            # 0ab 1ba ba
            if T[0][0] <= I[0][a] and T[1][1] <= I[1][b] < I[1][a] and I[2][b] < I[2][a]:
                return True
        P = P[1], P[2], P[0]  # circular permutation
        T = T[1], T[2], T[0]
        I = I[1], I[2], I[0]
    return False


def contains_forbidden3_coded(perm_triplet, tag_triplet):
    """Check whether the tagged triplet contains a minimal forbidden pattern of length 3"""
    if len(perm_triplet[0]) < 3:
        return False
    n = len(perm_triplet[0])
    P, T = perm_triplet, tag_triplet
    I = [inverse_perm(p) for p in P]
    for _ in range(3):
        for i, j, k in combinations(range(len(P[0])), 3):
            a, b, c = i, j, k
            if I[0][a] > I[0][b]:
                a, b = b, a
            if I[0][b] > I[0][c]:
                if I[0][a] > I[0][c]:
                    a, b, c = c, a, b
                else:
                    b, c = c, b
            # now a,b,c are sorted following P[0]
            # abc0 y1=x0 {ab}0 (first pattern rotated) => here x0 is 'a'
            if I[0][c] < T[0][0] and (I[1][b] < I[1][a] < I[1][c] or I[1][c] < I[1][a] < I[1][b]) and (
                    I[2][a] < T[2][0] and I[2][b] < T[2][0]):
                return True
            # abc0  1{ab} z1=x0 (second pattern) => here x0 is 'a'
            if I[0][c] < T[0][0] and (I[2][b] < I[2][a] < I[2][c] or I[2][c] < I[2][a] < I[2][b]) and (
                    I[1][a] >= T[1][1] and I[1][b] >= T[1][1]):
                return True
            # 1abc y1=x2 {bc}0 (third pattern rotated and a<->c) => x2 is 'c'
            if I[0][a] >= T[0][1] and (I[1][a] < I[1][c] < I[1][b] or I[1][b] < I[1][c] < I[1][a]) and (
                    I[2][b] < T[2][0] and I[2][c] < T[2][0]):
                return True
            # 1abc  1{bc} z1=x2 (fourth pattern) => here x2 is 'c'
            if I[0][a] >= T[0][1] and (I[2][a] < I[2][c] < I[2][b] or I[2][b] < I[2][c] < I[2][a]) and (
                    I[1][b] >= T[1][1] and I[1][c] >= T[1][1]):
                return True
            # abc0 b1ac ca0b
            if I[0][c] < T[0][0] and I[1][b] < T[1][1] <= I[1][a] < I[1][c] and I[2][c] < I[2][a] < T[2][0] <= I[2][b]:
                return True
            # 1abc b1ca ac0b
            if I[0][a] >= T[0][1] and I[1][b] < T[1][1] <= I[1][c] < I[1][a] and I[2][a] < I[2][c] < T[2][0] <= I[2][b]:
                return True
        P = P[1], P[2], P[0]  # circular permutation
        T = T[1], T[2], T[0]
        I = I[1], I[2], I[0]
    return False


def contains_forbidden4_coded(perm_triplet, tag_triplet):
    """Check whether the tagged triplet contains a minimal forbidden pattern of length 4"""
    if len(perm_triplet[0]) < 4:
        return False
    n = len(perm_triplet[0])
    P, T = perm_triplet, tag_triplet
    I = [inverse_perm(p) for p in P]
    for _ in range(3):
        for i, j, k, l in combinations(range(len(P[0])), 4):
            a, b, c, d = i, j, k, l
            if I[0][a] > I[0][b]:
                a, b = b, a  # now a < b
            if I[0][c] > I[0][d]:
                c, d = d, c  # now c < d
            if I[0][a] > I[0][c]:
                a, c = c, a  # now a is the smallest
            if I[0][b] > I[0][d]:
                b, d = d, b  # now d is the greatest
            if I[0][b] > I[0][c]:
                b, c = c, b
            if I[1][b] < I[1][a] < I[1][d] < I[1][c]:  # y = badc
                if I[2][c] < I[2][d] < I[2][a] < I[2][b]:  # z = cdab
                    # 0b1 1d a0
                    if T[0][0] <= I[0][b] < T[0][1] and T[1][1] <= I[1][d] and I[2][a] < T[2][0]:
                        return True
                    # b0 1a a0
                    if I[0][b] < T[0][0] and T[1][1] <= I[1][a] and I[2][a] < T[2][0]:
                        return True
                if I[2][d] < I[2][c] < I[2][b] < I[2][a]:  # z = dcba
                    # b0c 0d1 1c
                    if I[0][b] < T[0][0] <= I[0][c] and T[1][0] <= I[1][d] < T[1][1] and T[2][1] <= I[2][c]:
                        return True
                    # c0 {0ab}{1dc} 1c
                    if I[0][c] < T[0][0] and T[2][1] <= I[2][c]:
                        left = max(I[1][a], I[1][b], T[1][0] - .5)
                        right = min(I[1][d], I[1][c], T[1][1] - .5)
                        if left <= right:
                            return True
        P = P[1], P[2], P[0]  # circular permutation
        T = T[1], T[2], T[0]
        I = I[1], I[2], I[0]
    return False


def compute_region(v, inv_triplet, tag_triplet):
    """Computes the region of a vertex v on its line:
    0 if negative, 1 if between 0 and 1, and 2 if greater than 1"""
    n = len(inv_triplet[0])
    c, i = v // n, v % n
    if inv_triplet[c][i] < tag_triplet[c][0]:
        return 0
    elif inv_triplet[c][i] < tag_triplet[c][1]:
        return 1
    else:
        return 2


def compute_base_graph(perm_triplet, tag_triplet):
    n = len(perm_triplet[0])
    B = [[False] * (3 * n) for _ in range(3 * n)]
    for l in range(3):  # range through the lines
        for i in range(tag_triplet[l][0] - 1):  # the part before 0 in perm number c
            # links of the form l[i] -> l[i+1] if both before 0, for l in {X,Y,Z}
            B[n * l + perm_triplet[l][i]][n * l + perm_triplet[l][i + 1]] = True
        for i in range(tag_triplet[l][0], tag_triplet[l][1] - 1):  # between 0 and 1
            B[n * l + perm_triplet[l][i]][n * l + perm_triplet[l][i + 1]] = True
        for i in range(tag_triplet[l][1], n - 1):  # after 1
            B[n * l + perm_triplet[l][i]][n * l + perm_triplet[l][i + 1]] = True
    return B


def associated_vertex(v, n):
    """computes the w such that (v-f(w)) appears in the sign vector factorization"""
    c, i = v // n, v % n
    d = (c + 2) % 3
    return d * n + i


def compute_initial_orientation(i, l, j, inv_triplet, regions, U):
    """Computes the orientation of one entry (Xi, Yi, Zi, lj) of the sign vector.
    If it is determined by the tagged permutations, returns sign, None where sign=+-1
    Otherwise, it is of the form sign*sign_of(v-f(w)),with sign=+-1 and returns sign, v;
    v is then added to U (if not already in U, it is a set), which collects the missing information"""
    n = len(inv_triplet[0])
    s, t, u = l, (l + 1) % 3, (l + 2) % 3  # the three axis in direct order, starting with c_j
    if inv_triplet[s][i] > inv_triplet[s][j]:  # (si - sj) > 0
        sign = 1
    else:
        sign = -1
    if regions[t * n + i] < 2:  # (ti - 1) < 0
        sign = -sign
    v, w = i + u * n, i + t * n  # encodings of ui and ti as vertex numbers
    r1, r2 = regions[v], (regions[w] + 1) % 3  # regions of v and f(w)
    if r1 == r2:  # v and f(w) are in the same region, sign unknown, we add it to U
        U.add(v)
        return sign, v
    if r1 > r2:  # ui-f(ti) > 0
        return sign, None
    return -sign, None


def insertion_sort(T):
    """Sorts T, returns (-1)^#swaps """
    sign = 1
    for i in range(1, len(T)):
        key = T[i]
        j = i - 1
        while j >= 0 and key < T[j]:
            T[j + 1] = T[j]
            j -= 1
            sign = -sign
        T[j + 1] = key
    return sign


def compute_last(O, M, inv_triplet, regions):
    """Try to compute the orientation of O with the information stored in M
    returns sign, None if successful where sign is the orientation
    returns sign, [v,w] otherwise, when it the orientation is sign*(v-f(w))"""
    n = len(M) // 3
    # we first sort the variables while keeping (-1)^swaps in the variable sign
    # this way, we are in one of the cases described in the article
    sign = insertion_sort(O)
    line, index = [x // n for x in O], [x % n for x in O]

    if line[0] == line[1] and line[2] == line[3]:  # case (xa-xb)(yc-yd) or alike
        if inv_triplet[line[0]][index[0]] < inv_triplet[line[1]][index[1]]:
            sign = -sign  # (xa-xb) < 0
        if inv_triplet[line[2]][index[2]] < inv_triplet[line[3]][index[3]]:
            sign = -sign  # (yc-yd) < 0
        return sign, None
    if line[0] == line[1]:  # case [Xa,Xb,Yc,Zd]
        if inv_triplet[line[0]][index[0]] < inv_triplet[line[1]][index[1]]:
            sign = -sign  # (Xa-Xb) < 0
        if regions[O[2]] < 2:
            sign = -sign  # (Yc-1) < 0
        r3 = regions[O[3]]
        r2 = (regions[O[2]] + 1) % 3
        if r3 < r2 or M[O[3]][O[2]]:  # Zd < f(Yc)
            return -sign, None
        elif r3 > r2 or M[O[2]][O[3]]:  # Zd > f(Yc)
            return sign, None
        else:
            return sign, [O[3], O[2]]
    if line[1] == line[2]:  # case [Xa,Yb,Yc,Zd]
        sign = -sign
        if inv_triplet[line[1]][index[1]] < inv_triplet[line[2]][index[2]]:
            sign = -sign  # (yb-yc) < 0
        if regions[O[3]] < 2:
            sign = -sign  # (zd-1) < 0
        r0 = regions[O[0]]
        r3 = (regions[O[3]] + 1) % 3
        if r0 < r3 or M[O[0]][O[3]]:  # Xa < f(Zd)
            return -sign, None
        elif r0 > r3 or M[O[3]][O[0]]:  # Xa > f(Zd)
            return sign, None
        else:
            return sign, [O[0], O[3]]
    if line[2] == line[3]:  # case [Xa,Yb,Zc,Zd]
        if inv_triplet[line[2]][index[2]] < inv_triplet[line[3]][index[3]]:
            sign = -sign  # (zc-zd) < 0
        if regions[O[0]] < 2:
            sign = -sign  # (xa-1) < 0
        r1 = regions[O[1]]
        r0 = (regions[O[0]] + 1) % 3
        if r1 < r0 or M[O[1]][O[0]]:  # Yb < f(Xa)
            return -sign, None
        elif r1 > r0 or M[O[0]][O[1]]:  # Yb > f(Xa)
            return sign, None
        else:
            return sign, [O[1], O[0]]


def compute_sign_vector(sv_with_unknowns, M):
    """Computes the sign vector, once M contains all the needed information"""
    sv = list()
    n = len(M) // 3
    for k in range(6):
        if sv_with_unknowns[k][1] is None:
            sv.append(sv_with_unknowns[k][0])
        else:
            v = sv_with_unknowns[k][1]
            w = associated_vertex(v, n)
            if M[w][v]:  # v-f(w) > 0
                sv.append(sv_with_unknowns[k][0])
            else:
                sv.append(-sv_with_unknowns[k][0])
    return sv


def output_realization(L, regions, n):
    """Computes coordinates that match the order L and the repartition in regions
    returns an array of fractions C, where C[v] is the value of v on its line (as a fraction)"""
    C = [None] * (3 * n)
    i = 2
    for v in L:
        if regions[v] == 2:
            C[v] = Fraction(i)
        elif regions[v] == 0:
            C[v] = Fraction(-1, i - 1)
        else:
            C[v] = Fraction(i - 1, i)
        i += 1
    return C


# graph algorithms

def transitive_acyclic(M):
    """Computes the transitive closure of M, halts and return False if it is not acyclic"""
    for k in range(len(M)):
        for i in range(len(M)):
            for j in range(len(M)):
                M[i][j] = M[i][j] or (M[i][k] and M[k][j])
            if M[i][i]:
                return False
    return True


def add_edge_closure(M, v, w):
    """Add an edge v->w and update the transitive closure"""
    for i in range(len(M)):
        if i == v or M[i][v]:  # path from i to v
            for j in range(len(M)):
                if j == w or M[w][j]:  # path from w to j
                    M[i][j] = True


def topological_sort(M):
    """Computes a compatible order, M must be acyclic"""
    visited = [False] * len(M)
    L = []
    for v in range(len(M)):
        topological_rec(M, v, visited, L)
    L.reverse()
    return L


def topological_rec(M, v, visited, L):
    if visited[v]:
        return
    visited[v] = True
    for w in range(len(M)):
        if M[v][w]:
            topological_rec(M, w, visited, L)
    L.append(v)


# permutation algorithms

def mul_perm(p, q):
    """Returns the product pq of the two perms p and q"""
    return [p[q[i]] for i in range(len(p))]


def inverse_perm(p):
    """Returns the inverse permutation of p"""
    q = [0] * len(p)
    for i in range(len(p)):
        q[p[i]] = i
    return q

